#!/bin/bash
# Generated by runner ({{ runnerVersion }}) at {{ timestamp }}
set -e

{% set instance_data = jsonMap(container.instanceData()) %}
{% set cluster_data = jsonMap(container.clusterData()) %}
{% set esMajorVersion = version.major() %}
{% set esMinorVersion = version.minor() %}
{% set esPatchVersion = version.patch() %}

INSTALL='/elasticsearch/bin/elasticsearch-plugin install --silent --batch'

if [[ $(id -u) -eq 0 ]]; then
    echo "ERROR: elasticsearch.sh running as uid 0 (root), exiting."
    exit 1
fi

export ES_PATH_CONF="${ES_PATH_CONF:-/app/config}"
export CONF_DIR=$ES_PATH_CONF # bwc for 1.x and bin/elasticsearch-plugin on 5.x
export LOG_DIR="/app/logs"
export DATA_DIR="/app/data"
export USER_BUNDLES_DIR="/app/found_user_bundles"
export USER_BUNDLES_SCRIPTS_DIR="$USER_BUNDLES_DIR/scripts"
export USER_BUNDLES_DICTS_DIR="$USER_BUNDLES_DIR/dictionaries"

cd /app

mkdir -p logs
BOOT_LOG_FILE=logs/boot.log
# save stdout and stderr for later
exec 3>&1 4>&2
exec >  >(ts '%Y-%m-%dT%H:%M:%S%z' |tee -a $BOOT_LOG_FILE)
exec 2> >(ts '%Y-%m-%dT%H:%M:%S%z' |tee -a $BOOT_LOG_FILE >&2)

echo "Booting at $(date)"

mkdir -p plugins

# Without this setting, glibc will base the number of malloc arenas on the number of cores, which can hurt small processes running on large servers.
# Heroku has done a decent job benchmarking the effects of this setting, weighing performance of mallocs and actual memory used.
# A setting of 4 seems like a good tradeoff for our needs. https://devcenter.heroku.com/articles/testing-cedar-14-memory-use
export MALLOC_ARENA_MAX=4

# Number of GC threads should depend on the size of the node.
# Tiered compilation does not help Elasticsearch much, and reserves 48M memory for additional code caches.
# Disabling it reduces the code cache, however, which is problematic: https://github.com/elastic/cloud/issues/4603
export ES_JAVA_OPTS="-XX:ParallelGCThreads={{ container.parallelGCThreads() }} -XX:ConcGCThreads={{ container.concGCThreads() }}"

{% if enforceV4|default(false) %}
echo "Enforcing SigV4"
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dcom.amazonaws.services.s3.enforceV4"
{% endif %}

export ES_JAVA_OPTS="$ES_JAVA_OPTS -Xms{{ container.heapSize() }}M -Xmx{{ container.heapSize() }}M"

{% if container.heapSize() < 2048 %}
    # The heap isn't big, so switch to unpooled allocator in netty to save some heap
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dio.netty.allocator.type=unpooled"
{% endif %}

# Tuning for 5.x per https://github.com/elastic/cloud/issues/6153
# NIO poorly handles freeing direct buffers via the garbage collector.  Fixed in 5.5.1+
{% if esMajorVersion == 5 or esMajorVersion == 2 %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:+ExplicitGCInvokesConcurrent"
    {% if esMajorVersion == 2 or esMinorVersion <= 4 or (esMinorVersion == 5 and esPatchVersion == 0) %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:-DisableExplicitGC"
    {% endif %}
{% endif %}

# Use quota aware fs if included in this image.
if [ -f /elasticsearch/lib/quota-aware-fs-*.jar ]; then
    echo "Enabling QuotaAwareFileSystemProvider"
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -Djava.nio.file.spi.DefaultFileSystemProvider=co.elastic.cloud.quotaawarefs.QuotaAwareFileSystemProvider"
fi

# 5.0+ handles OOMs itself https://github.com/elastic/elasticsearch/pull/19272

{% if heapDumpOnOOM|default(false) %}
if [ "{{heapDumpEnabledUntil}}" -gt "$(date +%s)" ]; then
    echo "Enabling heap dump on oom"
    export ES_JAVA_OPTS="$ES_JAVA_OPTS -XX:+HeapDumpOnOutOfMemoryError -XX:HeapDumpPath=/app/heap_dumps"
fi
{% endif %}

# Less verbose curator
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dcurator-log-only-first-connection-issue-as-error-level=true"

# In 5.x the ability to set plugin.path was removed
# https://www.elastic.co/guide/en/elasticsearch/reference/5.0/breaking_50_plugins.html#_custom_plugins_path
# We maintain a link /elasticsearch/plugins -> /app/plugins for backward compatibility
# ref: https://github.com/elastic/elasticsearch/pull/22452
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dio.netty.recycler.maxCapacityPerThread=0"
export PLUGIN_TARGET_DIR="/elasticsearch/plugins"
export PLUGIN_SOURCE_DIR="/elasticsearch/plugin-archives"

{% if (esMajorVersion == 5 and ((esMinorVersion == 1 and esPatchVersion >= 2) or (esMinorVersion >= 2))) or (esMajorVersion >= 6) %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Des.cgroups.hierarchy.override=/"
{% endif %}


# In case of in-place version upgrades, make sure we clean up the plugins directory
# if this is a different version
{% set plugin_set_identifier = pluginSetIdentifier  | replace('"', "") %}
export PLUGIN_SET_MARKER="/data/found_installed_plugins-{{ plugin_set_identifier }}"
if [ ! -e $PLUGIN_SET_MARKER ]; then
    for PLUGIN in $(ls $PLUGIN_TARGET_DIR); do
        echo "Deleting $PLUGIN"
        rm -rf $PLUGIN_TARGET_DIR/$PLUGIN
        rm -rf /elasticsearch/bin/$PLUGIN
        # Only purge config if plugin will not be reinstalled, so that config provided by allocator is not deleted.
        if [[ "{{ instance_data.elasticsearch.plugins }}" =~ $PLUGIN ]]; then
            echo "Leaving config for $PLUGIN in place"
        else
            rm -rf /app/config/$PLUGIN
        fi
    done
    # After cleaning plugins, delete any marker file
    rm -f /app/found_installed_plugins-*
fi

{% for plugin_name in instance_data.elasticsearch.plugins %}
    {% set clean_plugin_name = plugin_name  | replace("'", "") %}
    if [ ! -e $PLUGIN_TARGET_DIR/{{ clean_plugin_name }} ]; then
        $INSTALL {{ name }} file://$(readlink -f $PLUGIN_SOURCE_DIR/{{ clean_plugin_name }}-*.zip)
    fi
{%  endfor %}

echo "Installing user plugins."
{% for user_plugin in instance_data.elasticsearch.user_plugins %}
    {% set name = user_plugin.name  | replace("'", "")  %}
    {% set url = user_plugin.url  | replace("'", "") %}
    {% if url != user_plugin.url %}
      echo "Single quotes were replaced from the plugin's URL, this might result in an error when starting the container"
    {% endif %}
    # installing user plugin: {{ name }}
    if [ ! -e $PLUGIN_TARGET_DIR/{{ name }} ] && [ ! -e $PLUGIN_TARGET_DIR/{{ name }}/.found_installed_plugin ]; then
        echo "Installing {{ name }}..."
        rm -rf $PLUGIN_TARGET_DIR/{{ name }} || true

            $INSTALL '{{ url }}' && touch $PLUGIN_TARGET_DIR/{{ name }}/.found_installed_plugin || true

        echo "Installed {{ name }}."
    else
        if [ -e $PLUGIN_TARGET_DIR/{{ name }}/.found_installed_plugin ]; then
            echo "Already installed {{ name }}."
        else
            echo "Not installing {{ name }} because it already exists as a built-in, enabled plugin"
        fi
    fi
{% endfor %}

# Indicate that plugins should be ok for this version, so we don't wipe everything above
touch $PLUGIN_SET_MARKER

echo "Installing user bundles."

mkdir -p "$USER_BUNDLES_DIR/.found-cached-downloads"
{% if not container.shouldInstallUserBundles() %}
echo "Skipping installation of user bundles due to allocation configuration."
{% elif not instance_data.elasticsearch.user_bundles %}
echo "No user bundles defined"
{% else %}
{% for user_bundle in instance_data.elasticsearch.user_bundles %}
    {% set name = user_bundle.name | replace("'", "") %}
    if [ ! -e $USER_BUNDLES_DIR/.found_installed_bundle_{{ name }} ]; then
        echo "Installing bundle [{{ name }}]..."
        {% set clean_bundle_url = user_bundle['url'] | replace("'", "") %}
        {% if clean_bundle_url != user_bundle['url'] %}
          echo "A single quote was filtered from the of the bundle. This might result on a container not starting, but could be a signal of an attack"
        {% endif %}
        {% set clean_bundle_url_hash = clean_bundle_url | md5 %}

        if [ ! -e $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip ]; then
          echo "Downloading bundle [{{ name }}]"
          if ! wget --no-use-server-timestamps -O $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip '{{ clean_bundle_url }}'; then
            echo "Could not download bundle [{{ name }}]"
            rm -f $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip
            exit 1
          fi

          echo "Validating bundle [{{ name }}]"
          if ! unzip -t $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip; then
            echo "Bundle [{{ name }}] is not a zip file"
            rm -f $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip
            exit 1
          fi

          echo "Clearing expired cached versions of bundle [{{ name }}]"
          ls -t $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-*.zip | tail -n +3 | xargs rm -f
        fi

        echo "Copying data from bundle [{{ name }}]"
        unzip -o $USER_BUNDLES_DIR/.found-cached-downloads/{{ name }}-{{ clean_bundle_url_hash }}.zip -d $USER_BUNDLES_DIR/

        echo "Installed bundle [{{ name }}]."
        echo "{{ clean_bundle_url_hash }}" > $USER_BUNDLES_DIR/.found_installed_bundle_{{ name }}
    else
        echo "Already installed bundle [{{ name }}]."
    fi
{% endfor %}

echo "Ensuring read/write permissions on all bundles"
chmod u+rw -R $USER_BUNDLES_DIR/*

{% if installBundlesFromStartScript|default(true) %}
# Copy the scripts directory to the found_user_bundles plugin
[[ -h $ES_PATH_CONF/scripts ]] && rm $ES_PATH_CONF/scripts
if [ -e $USER_BUNDLES_SCRIPTS_DIR ]; then
    cp -r $USER_BUNDLES_SCRIPTS_DIR $ES_PATH_CONF/scripts
fi

# Copy the dictionaries directory from the found_user_bundles
[[ -h $ES_PATH_CONF/dictionaries ]] && rm $ES_PATH_CONF/dictionaries
if [ -e $USER_BUNDLES_DICTS_DIR ]; then
    # We copy the dictionaries directly to /app/config. We need `-n` in order not to override existing files
    echo "Installing dictionaries"
    cp -rn $USER_BUNDLES_DICTS_DIR/* $ES_PATH_CONF
fi

# Maybe an old-style bundle
old_style_items=$(find $USER_BUNDLES_DIR -maxdepth 1 \
        -not -path $USER_BUNDLES_DIR \
        -not -path $USER_BUNDLES_SCRIPTS_DIR \
        -not -path $USER_BUNDLES_DICTS_DIR \
        -not -path "$USER_BUNDLES_DIR/\.found*")
for item in $old_style_items; do
    echo "Installing old-style-bundle dictionary: $item"
    cp -rn "$item" $ES_PATH_CONF
done
{% endif %}
{% endif %}

{% if esMajorVersion >= 5 %}
export ES_JAVA_OPTS="$ES_JAVA_OPTS -Dlog4j2.formatMsgNoLookups=true"
{% endif %}

echo "Done installing plugins and bundles, verifying required config files."

if [ ! -f /app/config/x-pack/system_key ]; then
    echo "Security System Key file not found (/app/config/x-pack/system_key)"
    exit 1
fi

echo "Done verifying required config files, starting Elasticsearch. See Elasticsearch logs for further output."

export CONFIG_OPTIONS="-Epath.logs=${LOG_DIR} -Epath.data=${DATA_DIR} -Epath.conf=${ES_PATH_CONF}"

# used by metricbeat
export ECE_COMPONENT="elasticsearch"
export ECE_CLUSTER_NAME="{{ container.id().group() }}"
export ECE_INSTANCE_ID="{{ container.id().name() }}"
export ECE_CLUSTER_VERSION="{{ container.versionString() }}"

# restore stdout and stderr
exec 1>&3 2>&4

# set cluster data environment variables, this is done to allow being able to
# modify the elasticsearch environment (for example java configuration and the like)
# without having to vacate the node, as Docker environment properties can't be modified
# while these properties are set by the allocator
{% for environment_variable in cluster_data.overrides.elasticsearchEnv %}
export {{environment_variable}}
{% endfor %}

# Do not abort script if elasticsearch returns error code
set +e
/elasticsearch/bin/elasticsearch -p /app/es.pid $CONFIG_OPTIONS $*
ES_EXIT_CODE=$?
if [ "$ES_EXIT_CODE" -ge "1" ]; then
    echo "Elasticsearch exited with status $ES_EXIT_CODE, running ./on-error-exitcode.sh $ES_EXIT_CODE"
    ./on-error-exitcode.sh $ES_EXIT_CODE
fi
exit $ES_EXIT_CODE
